# Copyright (C) 2001-2003 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# $Id: node.py,v 1.2 2003/05/29 00:31:40 halley Exp $

"""DNS nodes.  A node is a named set of rdatasets."""

import StringIO

import DNS.rdataset
import DNS.rdatatype

class Node(object):
    """A DNS node."""
    def __init__(self, name):
        self.name = name
        #
        # We use a list instead of a set so we can do the bad things (i.e.
        # having duplicate rdataset types) that dynamic update needs.
        #
        self.rdatasets = [];

    def to_text(self, **kw):
        s = StringIO.StringIO()
        for rds in self.rdatasets:
            print >> s, rds.to_text(self.name, **kw)
        return s.getvalue()[:-1]

    def __repr__(self):
        return '<DNS node ' + str(self.name) + '>'
    
    def __str__(self):
        return self.to_text()
    
    def __eq__(self, other):
        """Two nodes are equal if they have the same name and have the
        same rdatasets."""
        if self.name != other.name:
            return False
        #
        # This is inefficient.  Good thing we don't need to do it much.
        #
        for rd in self.rdatasets:
            if rd not in other.rdatasets:
                return False
        for rd in other.rdatasets:
            if rd not in self.rdatasets:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)
        
    def __len__(self):
        return len(self.rdatasets)

    def __getitem__(self, key):
        return self.rdatasets[key]

    def __setitem__(self, key, value):
        self.rdatasets[key] = value

    def __delitem__(self, key):
        del self.rdatasets[key]

    def find_rdataset(self, rdclass, rdtype, covers=DNS.rdatatype.NONE,
                      create=0, force_unique=0, deleting=0):
        """Find an rdataset matching the specified properties in the
        current node.  If create is non-zero, the rdataset is created if
        it doesn't exist, unless force_unique is also non-zero, in which
        case the rdataset will always be created."""
        if create and force_unique:
            rds = DNS.rdataset.Rdataset(rdclass, rdtype, deleting)
            self.rdatasets.append(rds)
            return rds
        for rds in self.rdatasets:
            if rds.match(rdclass, rdtype, covers, deleting):
                return rds
        if not create:
            raise KeyError
        rds = DNS.rdataset.Rdataset(rdclass, rdtype, deleting)
        self.rdatasets.append(rds)
        return rds

    def to_wire(self, file, compress=None, origin=None, question=False):
        count = 0
        for rds in self.rdatasets:
            count += rds.to_wire(self.name, file, compress, origin, question)
        return count
