# Copyright (C) 2003 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# $Id: resolver.py,v 1.5 2003/05/30 07:41:23 halley Exp $

"""DNS stub resolver."""

import socket
import sys
import time
import types

import DNS.message
import DNS.name
import DNS.query
import DNS.rcode
import DNS.rdataclass
import DNS.rdatatype

if sys.platform == 'win32':
    import _winreg

class NXDOMAIN(Exception):
    pass

class Timeout(Exception):
    pass

class NoAnswer(Exception):
    pass

class Answer(object):
    """DNS stub resolver answer

    Instances of this class bundle up the result of a successful DNS
    resolution.
    """
    def __init__(self, qname, rdtype, rdclass, response):
        self.qname = qname
        self.rdtype = rdtype
        self.rdclass = rdclass
        self.response = response
        for count in range(0, 15):
            node = None
            rds = None
            try:
                node = response.find_node(response.answer, qname)
                try:
                    rds = node.find_rdataset(rdclass, rdtype)
                    break
                except KeyError:
                    if rdtype != DNS.rdatatype.CNAME:
                        try:
                            rds = node.find_rdataset(rdclass,
                                                     DNS.rdatatype.CNAME)
                            for rd in rds:
                                qname = rd.target
                                break
                            continue
                        except KeyError:
                            raise NoAnswer
            except KeyError:
                raise NoAnswer
        if rds is None:
            raise NoAnswer
        self.node = node
        self.rdataset = rds

    def __getattr__(self, attr):
        if attr == 'name':
            return self.node.name
        elif attr == 'ttl':
            return self.rdataset.ttl
        elif attr == 'covers':
            return self.rdataset.covers
        else:
            raise AttributeError, attr

    def __len__(self):
        return len(self.rdataset)

    def __iter__(self):
        return iter(self.rdataset)

class Resolver(object):
    """DNS stub resolver
    """
    def __init__(self, filename='/etc/resolv.conf', configure=True):
        self.reset()
        if configure:
            if sys.platform == 'win32':
                self.read_registry()
            elif filename:
                self.read_resolv_conf(filename)

    def reset(self):
        """Reset all resolver configuration to the defaults."""
        self.domain = \
            DNS.name.Name(DNS.name.from_text(socket.gethostname())[1:])
        if len(self.domain) == 0:
            self.domain = DNS.name.root
        self.nameservers = []
        self.search = []
        self.port = 53
        self.timeout = 2.0
        self.lifetime = 30.0
        self.keyring = None
        self.keyname = None
        self.edns = -1
        self.ednsflags = 0
        self.payload = 0

    def read_resolv_conf(self, f):
        """Process f as a file in the /etc/resolv.conf format.  If f is
        a string, it is used as the name of the file to open; otherwise it
        is treated as the file itself."""
        if isinstance(f, str) or isinstance(f, unicode):
            f = open(f, 'r')
            want_close = True
        else:
            want_close = False
        try:
            for l in f:
                tokens = l.split()
                if tokens[0] == 'nameserver':
                    self.nameservers.append(tokens[1])
                elif tokens[0] == 'domain':
                    self.domain = DNS.name.from_text(tokens[1])
                elif tokens[0] == 'search':
                    for suffix in tokens[1:]:
                        self.search.append(DNS.name.from_text(suffix))
        finally:
            if want_close:
                f.close()

    def _config_win32_nameservers(self, nameservers):
        """Configure a NameServer registry entry."""
        # we call str() on nameservers to convert it from unicode to ascii
        ns_list = str(nameservers).split(',')
        for ns in ns_list:
            if not ns in self.nameservers:
                self.nameservers.append(ns)

    def _config_win32_domain(self, domain):
        """Configure a Domain registry entry."""
        # we call str() on domain to convert it from unicode to ascii
        self.domain = DNS.name.from_text(str(domain))

    def _config_win32_search(self, search):
        """Configure a Search registry entry."""
        # we call str() on search to convert it from unicode to ascii
        search_list = str(search).split(',')
        for s in search_list:
            if not s in self.search:
                self.search.append(DNS.name.from_text(s))

    def _config_win32_fromkey(self, key):
        """Extract DNS info from a registry key."""
        servers, rtype = _winreg.QueryValueEx(key, 'NameServer')
        if servers:
            self._config_win32_nameservers(servers)
            dom, rtype = _winreg.QueryValueEx(key, 'Domain')
            if dom:
                self._config_win32_domain(servers)
        else:
            servers, rtype = _winreg.QueryValueEx(key, 'DhcpNameServer')
            if servers:
                self._config_win32_nameservers(servers)
                dom, rtype = _winreg.QueryValueEx(key, 'DhcpDomain')
                if dom:
                    self._config_win32_domain(servers)
        search, rtype = _winreg.QueryValueEx(key, 'SearchList')
        if search:
            self._config_win32_search(servers)

    def read_registry(self):
        """Extract DNS info from the registry."""
        lm = _winreg.ConnectRegistry(None, _winreg.HKEY_LOCAL_MACHINE)
        want_scan = False
        try:
            try:
                # XP, 2000
                tcp_params = _winreg.OpenKey(lm,
                                             r'SYSTEM\CurrentControlSet'
                                             r'\Services\Tcpip\Parameters')
                want_scan = True
            except EnvironmentError:
                # ME
                tcp_params = _winreg.OpenKey(lm,
                                             r'SYSTEM\CurrentControlSet'
                                             r'\Services\VxD\MSTCP')
            try:
                self._config_win32_fromkey(tcp_params)
            finally:
                tcp_params.Close()
            if want_scan:
                interfaces = _winreg.OpenKey(lm,
                                             r'SYSTEM\CurrentControlSet'
                                             r'\Services\Tcpip\Parameters'
                                             r'\Interfaces')
                try:
                    i = 0
                    while True:
                        try:
                            guid = _winreg.EnumKey(interfaces, i)
                            i += 1
                            key = _winreg.OpenKey(interfaces, guid)
                            try:
                                # enabled interfaces seem to have a non-empty
                                # NTEContextList
                                (nte, ttype) = _winreg.QueryValueEx(key,
                                                             'NTEContextList')
                                if nte:
                                    self._config_win32_fromkey(key)
                            finally:
                                key.Close()
                        except EnvironmentError:
                            break
                finally:
                    interfaces.Close()
        finally:
            lm.Close()

    def query(self, qname, rdtype=DNS.rdatatype.A, rdclass=DNS.rdataclass.IN,
              tcp=False):
        if isinstance(qname, str):
            qname = DNS.name.from_text(qname, None)
        if isinstance(rdtype, str):
            rdtype = DNS.rdatatype.from_text(rdtype)
        if isinstance(rdclass, str):
            rdclass = DNS.rdataclass.from_text(rdclass)
        qnames_to_try = []
        if qname.is_absolute():
            qnames_to_try.append(qname)
        else:
            if len(qname) > 1:
                qnames_to_try.append(qname.concatenate(DNS.name.root))
            if self.search:
                for suffix in self.search:
                    qnames_to_try.append(qname.concatenate(suffix))
            else:
                qnames_to_try.append(qname.concatenate(self.domain))
        all_nxdomain = True
        start = time.time()
        for qname in qnames_to_try:
            request = DNS.message.make_query(qname, rdtype, rdclass)
            if not self.keyname is None:
                request.use_tsig(self.keyring, self.keyname)
            request.use_edns(self.edns, self.ednsflags, self.payload)
            response = None
            while response is None:
                for nameserver in self.nameservers:
                    now = time.time()
                    if now < start or now - start >= self.lifetime:
                        raise Timeout
                    try:
                        if tcp:
                            response = DNS.query.tcp(request, nameserver,
                                                     self.timeout, self.port)
                        else:
                            response = DNS.query.udp(request, nameserver,
                                                     self.timeout, self.port)
                    except:
                        response = None
                        continue
                    rcode = response.rcode()
                    if rcode == DNS.rcode.NOERROR or \
                           rcode == DNS.rcode.NXDOMAIN:
                        break
                    response = None
            if response.rcode() == DNS.rcode.NXDOMAIN:
                continue
            all_nxdomain = False
            break
        if all_nxdomain:
            raise NXDOMAIN
        return Answer(qname, rdtype, rdclass, response)

    def use_tsig(self, keyring, keyname=None):
        """When sending, a TSIG signature using the specified keyring and
        keyname will be added.  If the keyname is left unspecified, the
        first key in the keyring will be used."""
        self.keyring = keyring
        if keyname is None:
            self.keyname = self.keyring.keys()[0]
        else:
            self.keyname = keyname

    def use_edns(self, edns, ednsflags, payload):
        """Configure EDNS behavior.  The edns parameter is the EDNS level
        to use.  Specifying None or -1 means 'do not use EDNS'."""
        if edns is None:
            edns = -1
        self.edns = edns
        self.ednsflags = ednsflags
        self.payload = payload


def new(*args, **kw):
    return Resolver(*args, **kw)

default_resolver = None

def query(qname, rdtype=DNS.rdatatype.A, rdclass=DNS.rdataclass.IN,
          tcp=False):
    global default_resolver
    if default_resolver is None:
        default_resolver = new()
    return default_resolver.query(qname, rdtype, rdclass, tcp)
